 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */
package Java2OWL;

import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;
import java.util.HashMap;
import java.util.Map;
import org.apache.bcel.Constants;
import org.apache.bcel.classfile.Attribute;
import org.apache.bcel.classfile.AttributeReader;
import org.apache.bcel.classfile.ConstantObject;
import org.apache.bcel.classfile.ConstantPool;


/** This class  implements a reader for user defined annotations. It fills a gap in the BCEL-library.
 * <p>In contrast to Java reflection, where user defined annotations are accessible as Java objects,
 * BCEL extracts user defined annotations, but does not parse them. 
 * This must be done by the AnnotationReader. 
 * <p>This is adapted from an implementation by Don Schwarz.
 * 
 * <p> The annotation reader is installed by calling
 * {@code Attribute.addAttributeReader("RuntimeVisibleAnnotations",new AnnotationReader());}
 * 
 * @author Don Schwarz 
 */
public class AnnotationReader implements AttributeReader {
    
    /** This method is called when a class file is parsed by the BCEL-parser. 
     * It parses RuntimeVisibleAnnotations into instances of {@link AnnotationsAttribute}.
     */
    @Override
    public Attribute createAttribute (int name_index, int length, DataInputStream in, ConstantPool cp) {
       try {
            byte[] bytes = new byte[length];
            in.readFully(bytes);
            in = new DataInputStream(new ByteArrayInputStream(bytes));
            short numAnnotations = in.readShort();
            Map[] a = new Map[numAnnotations];
            for (int i = 0; i < numAnnotations; i++) {a[i] = readAnnotation(in, cp);}
            return new AnnotationsAttribute(Constants.ATTR_UNKNOWN, name_index, length, cp, a, bytes);} 
        catch (Exception ex) {throw new RuntimeException(ex);}}

    /** This method reads a single annotation */
    public Map readAnnotation (DataInputStream in, ConstantPool cp) throws IOException {
	short typeIndex = in.readShort();
	String type = cp.constantToString(cp.getConstant(typeIndex));
	Map map = new HashMap();
        short numMVPairs = in.readShort();
	for (int i = 0; i < numMVPairs; i++) {
            short nameIndex = in.readShort();
            String name = cp.constantToString(cp.getConstant(nameIndex));
            map.put(name, readMemberValue(in, cp));}
        map.put("_type", type);
        return map;}
    
    /** This method reads the value of a key-value pair. */
    public Object readMemberValue (DataInputStream in, ConstantPool cp) throws IOException, UnsupportedOperationException {
        byte tag = in.readByte(); 
        switch(tag) {
		case 'B': case 'C': case 'D': case 'J':
		case 'F': case 'I': case 'S': case 'Z':
                    return ((ConstantObject)cp.getConstant(in.readShort())).getConstantValue(cp);

		case 's': return cp.constantToString(cp.getConstant(in.readShort()));
                default:  throw new UnsupportedOperationException("tag = " + tag);
        }}
}
