 /*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */
package Java2OWL;

import java.io.File;
import java.util.*;
import org.semanticweb.owlapi.model.*;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.Map.Entry;

/** This class is for translating Java-classes into OWL-classes.
 * The main-method allows for calling it with the necessary parameters.
 * There are a number of other methods which can be called to translate Java-classes.
 *
 * @author ohlbach
 */
public class J2OCompiler{
    /** The top manager */ 
    private J2OManager manager;

    /** an empty constructor, just for test purposes. */
    J2OCompiler() {}
    
    /** returns the top manager */
    J2OManager getJ2OManager() {return manager;} 
    
    /** returns the name of the ontology manager.*/
    public String getName() {return manager.getName();}
    
    // These are the keywords to be used in the main-method.
    private static final String backgroundOntologyKey    = "bgo";
    private static final String extensionOntologyFileKey = "exof"; 
    private static final String extensionOntologyIRIKey  = "exoi";
    private static final String synchronizeKey           = "sync";
    private static final String logKey                   = "log";
    private static final String debugKey                 = "debug";
    private static final String reasonerKey              = "reasoner";
            static final String defaultReasoner          = "HermiT";
    private static final String defaultExtensionOntologyIRI = "http://" + System.getenv("LOGNAME")+"/ExtensionOntology.owl";
 
    private static final boolean buffering               = true;
    private static final boolean monitorProgress         = false;
    private static final int     timeout                 = 10000;
    private static final boolean importBackgroundOntology = true;
    
    static final Set<String> postfixes = new HashSet<String>(); 
    static final Set<String> keys = new HashSet<String>(); 
    
    static {
        postfixes.add(".class");   // for searching .class files in directories.
        keys.add(backgroundOntologyKey);
        keys.add(extensionOntologyFileKey);  
        keys.add(extensionOntologyIRIKey);
        keys.add(synchronizeKey);
        keys.add(logKey);
        keys.add(reasonerKey);
        keys.add(debugKey);
    }
    
    
    private J2OOntologyManager  j2oOntologyManager;
    private J2OClassManager     j2oClassManager;

    J2OClassManager getJ2oClassManager() {
        return j2oClassManager;}

    J2OOntologyManager getJ2oOntologyManager() {
        return j2oOntologyManager;}
    
    private boolean insertIntoOntology;

    /** logs events and error messages. */
    private Logging logger;
    
    
    /** creates a compiler with the given class manager 
     * 
     * @param j2oClassManager a class manager
     */
    J2OCompiler(J2OManager manager) {
        this.manager = manager;
        this.j2oClassManager    = manager.getClassManager();
        this.j2oOntologyManager = manager.getOntologyManager();
        logger = manager.getLogger();
        logger.activate(true);
        if(manager.statusCompile()) {
            this.insertIntoOntology = true;
            //try{logger.setHandler(Logging.Destination.BUFFER, "");} 
            //catch(Exception ex){ex.printStackTrace();}
        }
        logger.installation(this);}
    
    
   /** This method creates for a Java class the corresponding OWL-Class.
     * 
     * @param jcl The Java class to be turned into the OWL class.
     * 
     * <p>Only classes annotated with J2OWLClass are compiled. 
     * Besides jcl (if annotated), all its annotated superclasses and implemented interfaces are compiled.
     * In addition all range types of the J2OWLProperty-annotated methods are also compiled.
     * Inner classes are also compiled.
     * 
     * <p>So far only the subclass-superclass relations are turned into changeAxioms. 
     * The methods are not touched, and the change axioms are not processed.
     * This means the extension ontology is not yet changed. 
     */
    void class2OWL(Class jcl, J2OSynchroniser synchronizer) {
        if(jcl == null || jcl.isPrimitive() || j2oClassManager.isMapped(jcl) || jcl.isInterface() ||
                (jcl.getDeclaringClass() != null && (jcl.getModifiers() & Modifier.STATIC) != Modifier.STATIC)) 
            return;
        Package pkg = jcl.getPackage();
        if(pkg != null && (pkg.getName().startsWith("java") || pkg.getName().startsWith("org"))) return; // built-in classes need not be translated
        if(j2oClassManager.isMapped(jcl)) return;
        class2OWL(jcl.getSuperclass(),synchronizer);
        if(jcl.getAnnotation(Java2OWL.J2OWLClass.class) == null) return;   // class ought not to be translated.
        try {jcl.getConstructor();} 
        catch (Exception ex) {
            logger.error("The class has no constructor without arguments", Logging.Origin.JAVA, jcl);
            return;}
        if(jcl.getDeclaringClass() != null && (jcl.getModifiers() & Modifier.STATIC) != Modifier.STATIC) {
            logger.error("Only static inner classes can be translated", Logging.Origin.JAVA, jcl);
            return;}
        logger.entering(jcl);
        if(insertIntoOntology) j2oOntologyManager.class2OWL(jcl);
        else {
            J2OWLClass j2owlannotation = (J2OWLClass)jcl.getAnnotation(J2OWLClass.class);
            String oclName = j2owlannotation.name();      
            if(oclName.isEmpty()) oclName = jcl.getSimpleName(); 
            ArrayList<OWLClass> ocls = j2oOntologyManager.getOWLClasses(oclName); 
            if(ocls == null || ocls.size() == 0) logger.error("Unknown OWL-Class", Logging.Origin.USAGE, oclName);
            else {
                if(ocls.size() > 1) {logger.warning("More than one matching OWL-class", Logging.Origin.JAVA, oclName,ocls);}
                j2oClassManager.addClassCorrespondence(jcl,ocls.get(0));}}
        for(Method method : jcl.getDeclaredMethods()){
            if(method.getAnnotation(J2OWLProperty.class) != null) 
                class2OWL(Utilities.getReturnClass(method),synchronizer);}
        
        for(Class innerClass : jcl.getDeclaredClasses()) {class2OWL(innerClass,synchronizer);}
        if(synchronizer != null) {
            try{synchronizer.injectSynchronizer(jcl.getName());}
            catch(Exception ex) {
                logger.error("Error when synchronizing class", Logging.Origin.USAGE, ex);}}
        logger.exiting(jcl);
        return;}
 
    
     
     /** This method just calls the class2OWL-method for each class */
     public void class2OWL(Collection<Class> classes) {
         for(Class jcl : classes) class2OWL(jcl,null);
     }
    
     /** This method just calls the class2OWL-method for each class */
     public void class2OWL(Class... classes) {
          for(Class jcl : classes) class2OWL(jcl,null);
     }
    
     
     
    /** This method generates for all annotated methods the corresponding OWL-axioms 
      * @return false if something goes wrong. 
      * <p> Notice that the methods of superclasses have also to be added!
      */
    private boolean methods2OWL() {
        for(Class jcl : j2oClassManager.getClasses()) {
            for(Method method : jcl.getMethods()){
                if(method.getAnnotation(J2OWLProperty.class) != null) 
                    try{j2oClassManager.addMethod(jcl,method);}
                    catch(J2OException ex){continue;}}}
        String errors = "";
        for(PropertyWrapper prw : j2oClassManager.getPRWs()) {
            String error = prw.getInconsistencies();
            if(error != null && !error.isEmpty()) errors += prw.getInconsistencies();}
        if(!errors.isEmpty()) {
            logger.error("Inconsistencies in the generated ontology", Logging.Origin.JAVA, errors);
            return false;}
        for(PropertyWrapper prw : j2oClassManager.getPRWs()) {
            OWLProperty property = j2oOntologyManager.makeOWLProperty(prw.getName(), prw.isDataProperty());
            prw.setProperty(property,j2oOntologyManager.propertyIsFromBackground(property));
            if(insertIntoOntology) {
                j2oOntologyManager.addOWLAxioms(prw.createOWLAxioms());
                for(Entry<Class,PropertyWrapperComponent> entry : prw.getPropertyWrapperComponents().entrySet()) 
                    j2oOntologyManager.addMethodAttribute(prw.getProperty(),entry.getKey(),entry.getValue().getter);}}
        j2oOntologyManager.addEquivalences();
        return true;}
    
    /** performs the final steps of the translation 
     * 
     * @param extensionOntologyFile the name of the file where the extension ontology is to be saved.
     * @return true if the operation succeeded.
     * <p>The following steps are preformed.
     * <ol>
     * <li> The annotated methods are translated into OWL-properties<br/>
     * The next steps are only performed if insertIntoOntology = true.
     * <li> The changes to the ontology are made effective
     * <li> Inconsistent classes are extracted
     * <li> If extensionOntologyFile != "" then the extension ontology is saved. 
     * </ol>
     */ 
    public boolean finishTranslation(IRI extensionOntologyIRI) {
        if(!methods2OWL()) return false;
        if(insertIntoOntology) {
            j2oOntologyManager.addJavaClassesAsAnnotation(j2oClassManager.getClasses());
            if(extensionOntologyIRI != null) {
                j2oOntologyManager.saveExtensionOntology(extensionOntologyIRI);}}
        return true;}
   
    /** performs the final steps of the translation 
     * 
     * @param extensionOntologyFile the name of the file where the extension ontology is to be saved.
     * @return true if the operation succeeded.
     * <p>The following steps are preformed.
     * <ol>
     * <li> The annotated methods are translated into OWL-properties<br/>
     * The next steps are only performed if insertIntoOntology = true.
     * <li> The changes to the ontology are made effective
     * <li> Inconsistent classes are extracted.
     * </ol>
     */ 
    public boolean finishTranslation() {
        if(!methods2OWL()) return false;
        if(insertIntoOntology) {
            j2oOntologyManager.addJavaClassesAsAnnotation(j2oClassManager.getClasses());
             j2oOntologyManager.saveExtensionOntology();}
        return true;}
   
    
    
    /** This method restructures the grouped parameters from the main-method.
     * 
     * @param parameters 1. the class-filenames, 2. key-value pairs. 3. the directories.
     * @return an array of objects with the restructured parameters. 
     * <p>
     * <table>
     * <tr><td>1. classes,                  </td><td>  a list of Java-classes.<td></tr>
     * <tr><td>2. backgroundOntologyIRI,    </td><td>  The IRI for the background ontology.<td></tr>
     * <tr><td>3. extensionOntologyFileIRI, </td><td>  The File-IRI for the extension ontology.<td></tr>
     * <tr><td>4. extensionOntologyIRIIRI,  </td><td>  The IRI-of the extension ontology<.td></tr>
     * <tr><td>5. reasonerName              </td><td>  The reasoner name.<td></tr>
     * <tr><td>6. synchronizeFiles          </td><td>  true if class files are to be translated.</tr>
     *</table>
     */ 
    Object[] parseParameters(Object[] parameters) {
        ArrayList<String> classnames = ((HashMap<String,ArrayList<String>>)(parameters[0])).get(".class");
        HashMap<String, String> keyValues = (HashMap<String, String>)(parameters[1]);
        ArrayList<String> directorynames = (ArrayList<String>)(parameters[2]);
        if(keyValues.get(logKey) == null) {logger.activateInfo(false);}
        String debugValue = keyValues.get(debugKey);
        boolean debugMode = (debugValue == null || debugValue.equals("true"));
        manager.setDebugMode(debugMode);
        
        ArrayList<Class> classes = new ArrayList<Class>();
        for(String classname : classnames) { 
            Class jcl = null;
            File file = null;
            try{file = new File(classname);
                jcl = Utilities.classFromFile(file);
                if(jcl != null) classes.add(jcl);
            }
            catch(Exception ex) {
                ex.printStackTrace();
                logger.error("Class Load Exception", Logging.Origin.USAGE, classname, ex.getMessage());}
            catch(Error ex) {logger.error("Class Load Exception", Logging.Origin.USAGE, classname, ex.getMessage());}
            }
       
        for(String directoryname : directorynames) {
            File directory = new File(directoryname);
            if(directory.isDirectory()) { 
                Object[] result = Utilities.classesFromDirectory(directory);
                classes.addAll((ArrayList<Class>)(result[0]));
                if(!((String)(result[1])).isEmpty())
                    logger.error((String)(result[1]), Logging.Origin.USAGE);}
            else logger.error("The name is no directory", Logging.Origin.USAGE, directoryname);}
        
         String backgroundOntologyName = keyValues.get(backgroundOntologyKey);
         IRI backgroundOntologyIRI = null;
         if(!(backgroundOntologyName == null || backgroundOntologyName.isEmpty())) {
            if(backgroundOntologyName.startsWith("http:") || backgroundOntologyName.startsWith("file:"))
                backgroundOntologyIRI = IRI.create(backgroundOntologyName);
            else {
                File file = new File(backgroundOntologyName);
                if(!file.isFile()) logger.error("Background ontology is not a file",Logging.Origin.USAGE,backgroundOntologyName);
                else {
                    if(file.canRead()) backgroundOntologyIRI = IRI.create(file);
                    else logger.error("The background ontology is not readable",Logging.Origin.USAGE,backgroundOntologyName);}}}
         
         String extensionOntologyName = keyValues.get(extensionOntologyFileKey);
         File extensionOntologyFile = null;
         IRI extensionOntologyFileIRI = null;
         if(extensionOntologyName == null || extensionOntologyName.isEmpty()) {
             logger.error("No extension ontology filename provided",Logging.Origin.USAGE);}
         else {
             extensionOntologyFile = new File(extensionOntologyName);
             if(extensionOntologyFile.isDirectory()) {
                    logger.error("The extension ontology file is a directory", Logging.Origin.USAGE, extensionOntologyName);}
            extensionOntologyFileIRI = IRI.create(extensionOntologyFile);}
         
         extensionOntologyName = keyValues.get(extensionOntologyIRIKey);
         IRI extensionOntologyIRIIRI = null;
         if(extensionOntologyName == null || extensionOntologyName.isEmpty()) {
             if(backgroundOntologyIRI == null) {
                 extensionOntologyName = defaultExtensionOntologyIRI;
             }
             else {
                 String bgo = backgroundOntologyIRI.toString();
                 if(bgo.endsWith(".owl")) {
                     extensionOntologyName = bgo.substring(0,bgo.length()-4)+"Extension.owl";}
                 else extensionOntologyName = bgo+"extension.owl";}}
         extensionOntologyIRIIRI = IRI.create(extensionOntologyName);
         
         String reasonerName = keyValues.get(reasonerKey);
         if(reasonerName == null || reasonerName.isEmpty()) reasonerName = defaultReasoner;
         else {
             if(!J2OOntologyManager.getReasoners().contains(reasonerName)) {
                 logger.error("Unknown reasoner ",Logging.Origin.USAGE, reasonerName);}}
         
         String synchronizeValue = keyValues.get(synchronizeKey);
        
         boolean synchronizeFiles = (synchronizeValue != null && synchronizeValue.equals("true"));
         
         return new Object[]{classes, backgroundOntologyIRI, 
                             extensionOntologyFileIRI,extensionOntologyIRIIRI,
                             reasonerName, synchronizeFiles};
    }
    
    /** This method performs all the compilations.
     * 
     * @param parameters is the output of parseParameters
     * @return an error string if necessary.
     * <p>All classes are compiled and the extension ontology is saved.
     */
    void compile(Object[] parameters) {
        ArrayList<Class> classes = (ArrayList<Class>)(parameters[0]);
        IRI backgroundOntologyIRI = (IRI)(parameters[1]);
        IRI extensionOntologyFileIRI = (IRI)(parameters[2]);
        IRI extensionOntologyIRIIRI = (IRI)(parameters[3]);
        String reasonerName = (String)(parameters[4]);
        boolean synchronizeFiles = (Boolean)(parameters[5]);

        if(backgroundOntologyIRI != null) {
            j2oOntologyManager.loadOntology(backgroundOntologyIRI, true);}
        
        j2oOntologyManager.createExtensionOntology(importBackgroundOntology, extensionOntologyIRIIRI);
        
        j2oOntologyManager.setReasoner(reasonerName, buffering, monitorProgress, timeout);
       
        J2OSynchroniser synchronizer = null;
        if(synchronizeFiles) synchronizer = new J2OSynchroniser();
        for(int i = 0; i < classes.size(); ++i) {class2OWL(classes.get(i),synchronizer);}
        
        finishTranslation(extensionOntologyFileIRI);
        
        
    }
        
    /** This method turns the parameters generated by parseParameters into a string (for test purposes). */
    static String parametersToString(Object[] parameters) {
        ArrayList<Class> classes = (ArrayList<Class>)(parameters[0]);
        IRI backgroundOntologyIRI = (IRI)(parameters[1]);
        IRI extensionOntologyFileIRI = (IRI)(parameters[2]);
        IRI extensionOntologyIRIIRI = (IRI)(parameters[3]);
        String reasonerName = (String)(parameters[4]);
        boolean synchronizeFiles = (Boolean)(parameters[5]);
        String s = "";
        if(!classes.isEmpty()) {
            s += "Classes = ";
            for(int i = 0; i < classes.size(); ++i) {
                s += classes.get(i).getName() + ", ";}
            s += "\n";}
        if(backgroundOntologyIRI != null) s += "Background-Ontology IRI = " + backgroundOntologyIRI.toString() + "\n";
        if(extensionOntologyFileIRI != null) s += "Extension-Ontology File IRI = " + extensionOntologyFileIRI.toString() + "\n";
        if(extensionOntologyIRIIRI != null) s += "Extension-Ontology IRI IRI = " + extensionOntologyIRIIRI.toString() + "\n";
        if(reasonerName != null) s += "Reasoner name = " + reasonerName + "\n";
        s += "SynchronizeFiles = " + synchronizeFiles+ "\n"; 
        return Utilities.adjustPairs(s, "=");}
            
        
        
          
    /** This method returns a help-text for the main-method. */
    static String getHelptext() {
        String text = "The program can turn annotated compiled Java classes into an OWL ontology.\n";
        text += "The arguments to the program consist of filenames for Java .class-files,\n";
        text += "directorynames with Java .class-files and a number of key-value parameters.\n";
        text += "There can be any number of Java .class files and directories.\n";
        text += "The following key-value parameters should be provided:\n";
        
        text += "key = " + backgroundOntologyKey + "\n";
        text += "The value is the filename for the background ontology.\n";
        text += "It can be omitted if there is no background ontology.\n";
        
        text += "key = " + extensionOntologyFileKey + " (mandatory)\n";
        text += "The value is the filename where the newly created extension ontology is to be saved.\n";
        
        text += "key = " + extensionOntologyIRIKey + "\n";
        text += "The value is the IRI for the newly created extension ontology.\n";
        text += "By default some reasonable name is created from the background ontology or from the user name.\n";
        
        text += "key = " + synchronizeKey + "\n";
        text += "The value is true/false. If true, then synchronizer code is injected into the class files.\n";
        
        text += "key = " + debugKey + "\n";
        text += "The value is true/false. If true, then the system runs in debug mode. (default: true)\n";
        
        
        text += "key = " + reasonerKey + "\n";
        text += "The value is the name of the reasoner to be used.\n";
        text += "The following reasoners are available: " + J2OOntologyManager.getReasoners().toString() +".\n";
        text += "The default reasoner is " + defaultReasoner +"\n"; 
        
        text += "key = " + logKey + "\n";
        text += "The value 'all' causes all loggings to be printed, otherwise only error messages are printed.\n";
        
        return text;
    }
    
    /** This method reads all necessary arguments and performs all the translations. */
    public static void main(String[] args) {
        if(args.length == 0 || args[0].equals("help")) {System.out.println(getHelptext()); return;}
        J2OManager manager = new J2OManager("J2OManager",J2OManager.Status.COMPILE,true);
        J2OCompiler compiler = manager.getCompiler();
        
        Object[] parameters = Utilities.parseParameters(args, postfixes, keys);
        parameters = compiler.parseParameters(parameters);
        compiler.compile(parameters);
        System.out.println("Translation succeeded");}


}
