/*
    * Copyright (c) 2012, H. J. Ohlbach. All rights reserved.
    * DO NOT ALTER OR REMOVE COPYRIGHT NOTICES OR THIS FILE HEADER.
    *
    * This code is free software; you can redistribute it and/or modify it
    * under the terms of the GNU General Public License version 2 only, as
    * published by the Free Software Foundation. 
    *
    * This code is distributed in the hope that it will be useful, but WITHOUT
    * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
    * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
    * version 2 for more details.
    */

package Java2OWL;

import java.lang.reflect.Method;
import java.lang.reflect.ParameterizedType;
import java.lang.reflect.Type;
import java.util.Collection;
import java.util.Map;

/** This class is responsible for choosing the right type of PropertyWrapper for a given getter method.
 *
 * @author ohlbach
 */
class PropertyWrapperFactory {
    
    /** constructs the right type of property wrapper
     * 
     * @param jcl       An annotated Java class.
     * @param getter    An annotated getter method
     * @param manager   the top manager.
     * @return          a new PropertyWrapper.
     * @throws J2OException if something goes wrong.
     */
    static PropertyWrapper makePropertyWrapper(Class jcl, Method getter, J2OManager manager) throws J2OException {
        if(getter == null) throw new J2OException("makePropertyWrapper: No getter provided");
        Type type = getter.getGenericReturnType();
        Class cl = getter.getReturnType();
        PropertyWrapper prw = null; 
        if(isMappingFunctional(cl,type))        prw = new PropertyWrapperMappingFunctional(jcl,getter,manager);
        else {if(isMappingRelational(cl,type))  prw = new PropertyWrapperMappingRelational(jcl,getter,manager);
            else {if(isRelational(cl))          prw = new PropertyWrapperRelational(jcl,getter,manager);
                else {if(isFunctional(cl,type)) prw = new PropertyWrapperFunctional(jcl,getter,manager);}}}
        if(prw == null) manager.getLogger().error("Illegal return type for getter",Logging.Origin.USAGE,jcl,getter);
   
        else { 
            String errors = prw.getInconsistencies();
            if(errors == null) return prw;
            errors = "Inconsistencies in property " + prw.getName() + " in class " + jcl.getName() + "\n" + errors;
            manager.getLogger().error(errors,Logging.Origin.USAGE);
            throw new J2OException(errors);}
        return null;}
    
    /** checks whether the given class is a functional mapping, something like HashMap<String,String>
     * <p>Map<String,String> is not allowed.
     * The Type variables are not checked!.
     */ 
    static private boolean isMappingFunctional(Class cl, Type type) {
        return Map.class.isAssignableFrom(cl) && 
                (Utilities.isClass(((ParameterizedType)type).getActualTypeArguments()[1]) != null);}
        
     /** checks whether the given class is a relational mapping, something like HashMap<String,List<String>>
     * <p>Map<String,List<String>> is not allowed.
     * The Type components are not checked!.
     */ 
    static private boolean isMappingRelational(Class cl, Type type) {
        return Map.class.isAssignableFrom(cl)  && 
                (Utilities.isClass(((ParameterizedType)type).getActualTypeArguments()[1]) == null);}
    
    /** checks whether the given class is a collection, something like List<String> or String[]
     * <p>The Type variables are not checked!.
     */ 
    static private boolean isRelational(Class cl) {
        return cl.isArray() || Collection.class.isAssignableFrom(cl);}
    
    /** checks whether the type and class are just ordinary classes, like String.
     */ 
    static private boolean isFunctional(Class cl, Type t) {
            return !cl.isArray() && t == cl;}
}
